#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  Yandex.Disk indicator v.1.1.6
#
#  Copyright 2014 Sly_tom_cat <slytomcat@mail.ru>
#  based on grive-tools (C) Christiaan Diedericks (www.thefanclub.co.za)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import AppIndicator3 as appindicator
from gi.repository import GdkPixbuf
from gi.repository import Notify
import os
import sys
import subprocess
import datetime
import shutil
import tempfile
import webbrowser
import pyinotify
import asyncore
import threading
import fcntl
import gettext
import locale

def debugPrint(textToPrint):
  global verboseDebug
  if verboseDebug:
    try:
      print('%s: %s'%(datetime.datetime.now().strftime("%M:%S.%f"), textToPrint))
    except:
      pass

def sendmessage(title, message):
  global notificationSetting
  global notifier
  if notificationSetting:
    debugPrint('Message :%s'%message)
    notifier.update(title, message, yandexDiskIcon)    # Update notification
    notifier.show()                                    # Display new notification

def copyFile(source, destination):
  try:
    shutil.copy (source, destination)
  except:
    debugPrint("File Copy Error: from %s to %s" %(source,destination))

def deleteFile(source):
  try:
    os.remove(source)
  except:
    debugPrint('File Deletion Error: %s' % source)

def stopTimer(timerName):
  if timerName > 0:
    GLib.source_remove(timerName)
  return

def appExit():
  lockFile.write('')
  fcntl.flock(lockFile, fcntl.LOCK_UN)
  lockFile.close()
  os._exit(0)

def quitApplication(widget):
  global iconAnimationTimer
  global asyncTimer
  global watchTimer
  try:
    stopOnExit = settings.get_boolean("stoponexit")
  except:
    stopOnExit = False
  debugPrint('Stop daemon on exit is - %s' % str(stopOnExit))
  if stopOnExit:
    # Stop daemon
    stopYDdaemon()
    debugPrint('Daemon is stopped')
  # Stop timers
  stopTimer(iconAnimationTimer)
  stopTimer(asyncTimer)
  stopTimer(watchTimer)
  debugPrint("Timers are closed")
  # Close Remaining Threads
  asyncore.close_all()
  debugPrint("Async loop thread is closed")
  appExit()

def openLast(widget, index):
  global yandexDiskFolder
  global pathsList
  path = pathsList[index]     # Get path by index
  if path != '...':
    openPath(widget, os.path.join(yandexDiskFolder,path))

def openPath(widget, path):
  debugPrint('Opening %s'%path)
  if os.path.exists(path):
    try:
      os.startfile(path)
    except:
      subprocess.Popen(['xdg-open', path])

def openInBrowser(widget, url):
  webbrowser.open_new(url)

def daemonErrorDialog(err):
  # Show error messages according to the error
  if err in 'NOCONFIG':
    dialog = Gtk.MessageDialog(None, 0, Gtk.MessageType.INFO, Gtk.ButtonsType.OK_CANCEL, _('Yandex.Disk Indicator: daemon start failed'))
    dialog.format_secondary_text(_('Yandex.Disk daemon failed to start because it is not configured properly\n  To configure it up: press OK button.\n  Press Cancel to exit.'))
  else:
    dialog = Gtk.MessageDialog(None, 0, Gtk.MessageType.INFO, Gtk.ButtonsType.OK, _('Yandex.Disk Indicator: daemon start failed'))
    if err == 'NONET':
      dialog.format_secondary_text(_('Yandex.Disk daemon failed to start due to network connection issue. \n  Check the Internet connection and try to start daemon again.'))
    else:      #err == 'NOTINSTALLED'
      dialog.format_secondary_text(_('Yandex.Disk utility is not installed.\n Visit www.yandex.ru, download and install Yandex.Disk daemon.'));
  dialog.set_default_size(400, 250)
  response = dialog.run()
  dialog.destroy()
  if (err == 'NOCONFIG') and (response == Gtk.ResponseType.OK):
    # Launch Set-up utility
    debugPrint('starting configuration utility: %s'% os.path.join(installDir,'ya-setup'))
    retCode = subprocess.call([os.path.join(installDir,'ya-setup')])
  else:
    retCode = 1
  return retCode    # 0 when configuration was done successfully

def openAbout(widget):
  widget.set_sensitive(False) # Disable menu item
  aboutWindow = Gtk.AboutDialog()
  aboutWindow.set_name(_('Yandex.Disk indicator'))
  aboutWindow.set_version(_('Version ')+appVersion)
  aboutWindow.set_copyright('Copyright '+u'\u00a9'+
  ' 2013-'+datetime.datetime.now().strftime("%Y")+'\nSly_tom_cat')
  aboutWindow.set_comments(_('Yandex.Disk indicator \n(Grive Tools was used as example)'))
  aboutWindow.set_license(''+
  'This program is free software: you can redistribute it and/or \n'+
  'modify it under the terms of the GNU General Public License as \n'+
  'published by the Free Software Foundation, either version 3 of \n'+
  'the License, or (at your option) any later version.\n\n'+
  'This program is distributed in the hope that it will be useful, \n'+
  'but WITHOUT ANY WARRANTY; without even the implied warranty \n'+
  'of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. \n'+
  'See the GNU General Public License for more details.\n\n'+
  'You should have received a copy of the GNU General Public License \n'+
  'along with this program.  If not, see http://www.gnu.org/licenses')
  aboutWindow.set_authors([_('Sly_tom_cat (slytomcat@mail.ru) '),
  _('ya-setup utility author: Snow Dimon (snowdimon.ru)' ), 
  _('\nSpecial thanks to:'),
  _(' - Christiaan Diedericks (www.thefanclub.co.za) - Grive tools creator'),
  _(' - ryukusu_luminarius (my-faios@ya.ru) - icons designer'),
  _(' - metallcorn (metallcorn@jabber.ru) - icons designer'),
  _(' - Chibiko (zenogears@jabber.ru) - deb package creation assistance'),
  _(' - RingOV (ringov@mail.ru) - localization assistance'),
  _(' - And to all other people who contributed to this project through'),
  _('   the Ubuntu.ru forum http://forum.ubuntu.ru/index.php?topic=241992)')] )
  aboutWindow.set_logo(GdkPixbuf.Pixbuf.new_from_file(yandexDiskIcon))
  aboutWindow.set_icon(GdkPixbuf.Pixbuf.new_from_file(yandexDiskIcon))
  aboutWindow.set_program_name(_('Yandex.Disk indicator'))
  aboutWindow.run()
  aboutWindow.destroy()
  widget.set_sensitive(True)  # Enable back menu item

def onCheckButtonToggled(widget, button, key):
  global notificationSetting
  global settings
  global overwrite_check_button
  # Update dconf settings
  toggleState = button.get_active()
  settings.set_boolean(key, toggleState)
  debugPrint('Togged: '+key+'  val: '+str(toggleState))
  if key == 'theme':
    updateIconTheme()        # Update themeStyle
    updateIcon()              # Update current icon
  elif key == 'notifications':
    # Update notifications variable
    notificationSetting = toggleState    
  elif key == 'autostartdaemon':
    if toggleState:
      copyFile(autoStartSource1, autoStartDestination1)
      sendmessage(_('Yandex.Disk daemon'), _('Auto-start ON'))
    else:
      deleteFile(autoStartDestination1)
      sendmessage(_('Yandex.Disk daemon'), _('Auto-start OFF'))
  elif key == 'autostart':
    if toggleState:
      copyFile(autoStartSource, autoStartDestination)
      sendmessage(_('Yandex.Disk Indicator'), _('Auto-start ON'))
    else:
      deleteFile(autoStartDestination)
      sendmessage(_('Yandex.Disk Indicator'), _('Auto-start OFF'))
  elif key == 'fmextensions':
    activateActions(toggleState)
  elif key == 'optionreadonly':
    overwrite_check_button.set_sensitive(toggleState)

def openPreferences(menu_widget):
  # Preferences Window
  global settings
  global overwrite_check_button
  # Disable menu item to avoid multiple preferences windows creation
  menu_widget.set_sensitive(False)
  # Create Preferences window
  preferencesWindow = Gtk.Dialog()
  preferencesWindow.set_title(_('Yandex.Disk-indicator and Yandex.Disk preferences'))
  preferencesWindow.set_icon(GdkPixbuf.Pixbuf.new_from_file(yandexDiskIcon))
  preferencesWindow.set_border_width(6)
  preferencesWindow.add_button(_('Close'),Gtk.ResponseType.CLOSE)
  # Create notebook for indicator and daemon options 
  pref_notebook = Gtk.Notebook()
  # Put it inside the dialog content area
  preferencesWindow.get_content_area().add(pref_notebook)
  # --- Indicator preferences tab ---  
  preferencesBox = Gtk.Box.new(Gtk.Orientation.VERTICAL,5)
  # Auto-start indicator on system start-up
  optionName = _('Start Yandex.Disk indicator when you start your computer')
  key = 'autostart'
  autostart_check_button = Gtk.CheckButton(optionName)
  autostart_check_button.set_active(settings.get_boolean(key))
  autostart_check_button.connect("toggled", onCheckButtonToggled, autostart_check_button, key)
  settings.bind(key, autostart_check_button, 'active', Gio.SettingsBindFlags.GET)
  preferencesBox.add(autostart_check_button)
  # Notifications
  optionName = _('Show on-screen notifications')
  key = 'notifications'
  notifications_check_button = Gtk.CheckButton(optionName)
  notifications_check_button.set_active(settings.get_boolean(key))
  notifications_check_button.connect("toggled", onCheckButtonToggled, notifications_check_button, key)
  settings.bind(key, notifications_check_button, 'active', Gio.SettingsBindFlags.GET)
  preferencesBox.add(notifications_check_button)
  # Theme
  optionName = _('Prefer light icon theme')
  key = 'theme'
  theme_check_button = Gtk.CheckButton(optionName)
  theme_check_button.set_active(settings.get_boolean(key))
  theme_check_button.connect("toggled", onCheckButtonToggled, theme_check_button, key)
  settings.bind(key, theme_check_button, 'active', Gio.SettingsBindFlags.GET)
  preferencesBox.add(theme_check_button)
  # Activate file-manager extensions
  optionName = _('Activate file manager extensions')
  key = 'fmextensions'
  fmext_check_button = Gtk.CheckButton(optionName)
  fmext_check_button.set_active(settings.get_boolean(key))
  fmext_check_button.connect("toggled", onCheckButtonToggled, fmext_check_button, key)
  settings.bind(key, fmext_check_button, 'active', Gio.SettingsBindFlags.GET)
  preferencesBox.add(fmext_check_button)
  # --- End tab --- add it to notebook
  pref_notebook.append_page(preferencesBox, Gtk.Label(_('Indicator settings')))
  # --- Daemon start options tab ---
  optionsBox = Gtk.Box.new(Gtk.Orientation.VERTICAL,5)
  # Auto-start daemon on system start-up
  optionName = _('Start Yandex.Disk daemon when you start your computer')
  key = 'autostartdaemon'
  autostart_d_check_button = Gtk.CheckButton(optionName)
  autostart_d_check_button.set_active(settings.get_boolean(key))
  autostart_d_check_button.connect("toggled", onCheckButtonToggled, autostart_d_check_button, key)
  settings.bind(key, autostart_d_check_button, 'active', Gio.SettingsBindFlags.GET)
  optionsBox.add(autostart_d_check_button)
  # Stop daemon on exit
  optionName = _('Stop Yandex.Disk daemon on closing of indicator')
  key = 'stoponexit'
  stop_check_button = Gtk.CheckButton(optionName)
  stop_check_button.set_active(settings.get_boolean(key))
  stop_check_button.connect("toggled", onCheckButtonToggled, stop_check_button, key)
  settings.bind(key, stop_check_button, 'active', Gio.SettingsBindFlags.GET)
  optionsBox.add(stop_check_button)
  # Option Read-Only
  optionName = _('Read-Only: Do not upload locally changed files to Yandex.Disk')
  key = 'optionreadonly'
  readOnly_check_button = Gtk.CheckButton(optionName)
  readOnly_check_button.set_tooltip_text(_("Locally changed files will be renamed if a newer version of this file appear in Yandex.Disk. \n NOTE! You have to reload daemon to activate this setting"))
  readOnly_check_button.set_active(settings.get_boolean(key))
  readOnly_check_button.connect("toggled", onCheckButtonToggled, readOnly_check_button, key)
  settings.bind(key, readOnly_check_button, 'active', Gio.SettingsBindFlags.GET)
  optionsBox.add(readOnly_check_button)
  # Option Overwrite
  optionName = _('Overwrite locally changed files by files from Yandex.Disk (in read-only mode)')
  key = 'optionoverwrite'
  overwrite_check_button = Gtk.CheckButton(optionName)
  overwrite_check_button.set_tooltip_text(_("Locally changed files will be overwritten if a newer version of this file appear in Yandex.Disk. \n NOTE! You have to reload daemon to activate this setting"))
  overwrite_check_button.set_active(settings.get_boolean(key))
  overwrite_check_button.set_sensitive(settings.get_boolean('optionreadonly'))
  overwrite_check_button.connect("toggled", onCheckButtonToggled, overwrite_check_button, key)
  settings.bind(key, overwrite_check_button, 'active', Gio.SettingsBindFlags.GET)
  optionsBox.add(overwrite_check_button)
  # --- End of tab --- add it to notebook
  pref_notebook.append_page(optionsBox, Gtk.Label(_('Daemon options')))
  # ------ Dialog construction is finished -------
  # Run it
  preferencesWindow.show_all()
  preferencesWindow.run()
  preferencesWindow.destroy()
  optionsSave()                     # Save daemon statrt options in config file
  menu_widget.set_sensitive(True)   # Enable menu item

def renderMenu():
  # Build menu and attach it to indicator
  global menu_status
  global menu_used
  global menu_free
  global menu_YD_daemon_stop
  global menu_YD_daemon_start
  global widgetsList
  global pathsList
  global yandexDiskSpace1
  global yandexDiskSpace2
  global yandexDiskStatus
  global yandexDiskFolder
  global currentStatus
  menu = Gtk.Menu()           # Create menu
  menu_status = Gtk.MenuItem()
  menu_status.set_use_underline(False)
  menu_status.set_sensitive(False)
  menu_status.set_label(yandexDiskStatus)
  menu.append(menu_status)
  menu_used = Gtk.MenuItem()
  menu_used.set_use_underline(False)
  menu_used.set_sensitive(False)
  menu_used.set_label(yandexDiskSpace1)
  menu.append(menu_used)
  menu_free = Gtk.MenuItem()
  menu_free.set_use_underline(False)
  menu_free.set_sensitive(False)
  menu_free.set_label(yandexDiskSpace2)
  menu.append(menu_free)
  # Sub-menu: list of last synchronized
  submenu_last = Gtk.Menu()
  widgetsList = []            # List of menu_last widgets to update their labels later
  pathsList = []              # List of files/folders in lastMenu items
  for i in range(0, 9):
    pathsList.append('...')
    widgetsList.append(Gtk.MenuItem.new_with_label('...'))
    widgetsList[i].set_use_underline(False)
    widgetsList[i].connect("activate", openLast, i)
    submenu_last.append(widgetsList[i])
  menu_last = Gtk.MenuItem(_('Last synchronized items'))
  menu_last.set_use_underline(False)
  menu_last.set_submenu(submenu_last)
  menu.append(menu_last)
  menu.append(Gtk.SeparatorMenuItem.new()) #-----separator--------
  menu_YD_daemon_start = Gtk.MenuItem(_('Start Yandex.Disk daemon'))
  menu_YD_daemon_start.set_use_underline(False)
  menu_YD_daemon_start.connect("activate", startDaemon)
  if currentStatus != 'none':
    menu_YD_daemon_start.set_sensitive(False)
  menu.append(menu_YD_daemon_start)
  menu_YD_daemon_stop = Gtk.MenuItem(_('Stop Yandex.Disk daemon'))
  menu_YD_daemon_stop.set_use_underline(False)
  menu_YD_daemon_stop.connect("activate", stopDaemon)
  if currentStatus == 'none':
    menu_YD_daemon_stop.set_sensitive(False)
  menu.append(menu_YD_daemon_stop)
  menu_open_YD_folder = Gtk.MenuItem(_('Open Yandex.Disk Folder'))
  menu_open_YD_folder.set_use_underline(False)
  menu_open_YD_folder.connect("activate", openPath, yandexDiskFolder)
  menu.append(menu_open_YD_folder)
  menu_open_YD_web = Gtk.MenuItem(_('Open Yandex.Disk on the web'))
  menu_open_YD_web.set_use_underline(False)
  menu_open_YD_web.connect("activate", openInBrowser, 'http://disk.yandex.ru')
  menu.append(menu_open_YD_web)
  menu.append(Gtk.SeparatorMenuItem.new()) #-----separator--------
  menu_preferences = Gtk.MenuItem(_('Preferences'))
  menu_preferences.set_use_underline(False)
  menu_preferences.connect("activate", openPreferences)
  menu.append(menu_preferences)
  menu_help = Gtk.MenuItem(_('Help'))
  menu_help.set_use_underline(False)
  menu_help.connect("activate", openInBrowser, 'http://help.yandex.ru/disk/')
  menu.append(menu_help)
  menu_about = Gtk.MenuItem(_('About'))
  menu_about.set_use_underline(False)
  menu_about.connect("activate", openAbout)
  menu.append(menu_about)
  menu.append(Gtk.SeparatorMenuItem.new()) #-----separator--------
  menu_quit = Gtk.MenuItem(_('Quit'))
  menu_quit.set_use_underline(False)
  menu_quit.connect("activate", quitApplication)
  menu.append(menu_quit)
  menu.show_all()

  ind.set_menu(menu)          # Attach menu to indicator

def startDaemon(widget):
  # Try to start yandex-disk daemon and change the menu items sensitivity 
  global menu_YD_daemon_start
  global menu_YD_daemon_stop
  err = startYDdaemon()
  if err == '':
    menu_YD_daemon_start.set_sensitive(False)
    menu_YD_daemon_stop.set_sensitive(True)
  else:
    daemonErrorDialog(err)
    menu_YD_daemon_start.set_sensitive(True)
    menu_YD_daemon_stop.set_sensitive(False)

def stopDaemon(widget):
  global menu_YD_daemon_start
  global menu_YD_daemon_stop
  stopYDdaemon()
  menu_YD_daemon_start.set_sensitive(True)
  menu_YD_daemon_stop.set_sensitive(False)

def checkDaemon():
  # Checks that daemon process is running and daemon's response is correct,
  # and kills the daemon when it is running but not responding (HARON_CASE).
  try:
    msg = subprocess.check_output(['pgrep', '-x', 'yandex-disk'],universal_newlines=True)[ :-1]
  except:
    debugPrint("yandex-disk daemon is not running")
    return False
  debugPrint('yandex-disk daemon PID is: %s'%msg)
  # Check for correct daemon response
  if getDaemonOutput():
    debugPrint('yandex-disk daemon is responding correctly.')
    return True
  debugPrint('yandex-disk daemon is NOT responding!')
  # As it is not responding - try to kill all of them
  try:
    subprocess.check_call(['killall', 'yandex-disk'])
    debugPrint('yandex-disk daemon killed')
  except:
    debugPrint('yandex-disk daemon kill error')
  return False

def getDaemonOutput():
  global daemonOutput
  try:
    daemonOutput = subprocess.check_output(['yandex-disk', 'status'],universal_newlines=True)
  except:
    daemonOutput = ''
  #debugPrint('output = %s'%daemonOutput)
  return (daemonOutput != '')

def checkDaemonStatus():
  # Get and parses the daemon output and find status values and e.t.c. in it.
  # Pre-formats status messages for menu
  global currentStatus
  global yandexDiskStatus
  global yandexDiskSpace1
  global yandexDiskSpace2
  global syncProgress
  global lastItems
  global daemonOutput
  getDaemonOutput()
  # Look for synchronization progress
  lastPos = 0
  startPos = daemonOutput.find('ync progress: ')
  if startPos > 0:
    startPos += 14  # 14 is a length of 'ync progress: ' string
    lastPos = daemonOutput.find('\n',startPos)
    syncProgress = daemonOutput[startPos: lastPos]
  else:
    syncProgress = ''
  if daemonOutput != '':
    # Look for current core status (it is always presented in output)
    startPos = daemonOutput.find('core status: ',lastPos)+13
    # len('core status: ') = 13
    lastPos = daemonOutput.find('\n',startPos)
    currentStatus = daemonOutput[startPos: lastPos]
  else:
    currentStatus = 'none'
  # Look for total Yandex.Disk size
  startPos = daemonOutput.find('Total: ',lastPos)
  if startPos > 0:
    startPos += 7   # len('Total: ') = 7
    lastPos = daemonOutput.find('\n',startPos)
    sTotal = daemonOutput[startPos: lastPos]
    ## If 'Total: ' was found then other information as also presented
    # Look for used Yandex.Disk size. It must be there
    startPos = daemonOutput.find('Used: ',lastPos)+6 # len('Used: ') = 6
    lastPos = daemonOutput.find('\n',startPos)
    sUsed = daemonOutput[startPos: lastPos]
    # Look for free size on Yandex.Disk. It must be there
    startPos = daemonOutput.find('Available: ',lastPos)+11 # len('Available: ') = 11
    lastPos = daemonOutput.find('\n',startPos)
    sAvailable = daemonOutput[startPos: lastPos]
    # Look for trash size of Yandex.Disk
    startPos = daemonOutput.find('Trash size: ',lastPos)+12 # len('Trash size: ') = 12
    lastPos = daemonOutput.find('\n',startPos)
    sTrash = daemonOutput[startPos: lastPos]
  else: # When there is no Total: it is also other sizes are not presented
    sTotal = '...'
    sUsed = '...'
    sAvailable = '...'
    sTrash = '...'
  # Look for last synchronized items list
  startPos = daemonOutput.find('Last synchronized',lastPos)
  if startPos > 0:
    startPos = daemonOutput.find('\n',startPos)+1 # skip one line
    lastItems = daemonOutput[startPos: ].decode('utf-8') # save the rest and decode it
    # Decode required for python 2.7 and not required for python 3.0 and above
    # We don't split last synchronized list on individual lines. 
    # It is easer to do it in the same loop where menu is being updated.
  else:
    lastItems = ''
  # Prepare and format information for menu
  yandexDiskStatus = _('Status: ') + (\
  _('Synchronized')         if currentStatus == 'idle' else \
  _('Sync.: ')+syncProgress if currentStatus == 'busy' else \
  _('Indexing')             if currentStatus == 'index' else \
  _('Paused')               if currentStatus == 'paused' else \
  _('Daemon stopped')       if currentStatus == 'none' else \
  _('Not connected')        if currentStatus == 'no internet access' else \
  _('Error')             ) #if currentStatus == 'error' or any other errors
  yandexDiskSpace1 = _('Used: ')+sUsed+'/'+sTotal
  yandexDiskSpace2 = _('Free: ')+sAvailable+_(', trash: ')+sTrash
  return

def getYDFolder():
  # Search in 'yandex-disk status' output for the Yandex.disk folder path
  global daemonOutput
  startPos = daemonOutput.find("Yandex.Disk directory: '") + 24
  # 24 is a length of "Yandex.Disk directory: '" string
  return daemonOutput[startPos : daemonOutput.find("'",startPos)]

def startYDdaemon():
  # Execute 'yandex-disk start' and return '' if success or error message if not... but sometime it starts successfully with error message
  try:
    msg = subprocess.check_output(['yandex-disk', 'start'],universal_newlines=True)
    debugPrint('Start success, message: %s'%msg)
    return ''
  except subprocess.CalledProcessError as e:
    debugPrint('Start failed:%s'%e.output)
    if e.output == '': # probably 'os: no file'
      return 'NOTINSTALLED'
    err = 'NONET' if e.output.find('Proxy')>0 else\
          'BADDAEMON' if e.output.find('daemon')>0 else\
          'NOCONFIG'
    return err

def stopYDdaemon():
  # Execute 'yandex-disk stop'
  try:
    msg = subprocess.check_output(['yandex-disk','stop'],universal_newlines=True)
  except:
    msg = ''
  return (msg != '')

def updateLastSyncList():
  # Update last synchronized list
  global widgetsList
  global pathsList
  global lastItems
  endPos = 0
  for i in range(0, 9):
    startPos = lastItems.find("'",endPos+2)+1
    endPos = lastItems.find("\n",startPos)-1
    if (startPos<1) or (endPos<0):
      lastLine = '...'
    else:
      lastLine = lastItems[startPos:endPos]
      #debugPrint('"%s"'%sLine)
      if lastLine =='':
        lastLine = '...'
    pathsList[i] = lastLine   # Store path/file name to have an ability to open it later
    if len(lastLine) > 50:    # Is path too long for menu label?
      # Shorten it: get 23 symbols from beginning, then '...', and 24 from end (50 total)
      lastLine = lastLine[:22]+'...'+lastLine[-25:]
    # Replace underscore to disable use_underline feature of GTK menu acceleration
    widgetsList[i].set_label(lastLine.replace('_',u"\u02CD"))
    # Change sensitivity of last menu item according to the path existence.
    if os.path.exists(os.path.join(yandexDiskFolder,pathsList[i])):
      widgetsList[i].set_sensitive(True)
    else:
      widgetsList[i].set_sensitive(False)

def joinTreads():
  # Join threads every second for 0.5 seconds and relax
  global loop_thread
  loop_thread.join(timeout = 0.5)
  return True

def handleEvent(triggeredByWatcher):
  # Main working routine: event handler function.
  # It react (icon change/messages) on status change and also update
  # status information in menu (status, sizes, last synchronized items).
  # It can be triggered by timer (triggeredByWatcher=False) or
  # by iNonify watcher (triggeredByWatcher=True)
  global currentStatus
  global newStatus
  global lastStatus
  global menu_status
  global menu_YD_daemon_stop
  global menu_YD_daemon_start
  global watchTimer
  global yandexDiskSpace1
  global yandexDiskSpace2
  global yandexDiskStatus
  global timerTriggeredCount
  # Get the latest status data
  checkDaemonStatus()
  # Update status and sizes in menu on any event
  menu_status.set_label(yandexDiskStatus)
  menu_used.set_label(yandexDiskSpace1)
  menu_free.set_label(yandexDiskSpace2)
  # Update last synchronized list on any event
  updateLastSyncList()
  # Convert status to the internal presentation ['busy','idle','paused','none','error']
  newStatus = currentStatus
  if newStatus == 'index':
    # Don't handle 'index' status
    newStatus = lastStatus         # assume that status not changed
  # Convert all error statuses to one: 'error'
  elif newStatus not in ['busy','idle','paused','none']: #it is also != 'index'
    # Status in ['error', 'no internet access','failed to connect to daemon process',...]
    newStatus = 'error'

  debugPrint('Event: triggered by %s    status: %s -> %s' \
            % ('watcher'if triggeredByWatcher else 'timer  ',lastStatus, newStatus))

  if  lastStatus != newStatus:     # Has status changed?
    # Update current icon
    updateIcon()
    # React on change of status
    if newStatus == 'busy':
      # Just entered into 'busy' from some another status
      sendmessage(_('Yandex.Disk'), _('Synchronization is started'))
    elif newStatus == 'idle':
      # Just entered into 'idle' from some another status
      # Show message when status changed from 'busy' to 'idle'
      if lastStatus == 'busy':
        sendmessage(_('Yandex.Disk'), _('Synchronization is finished'))
    elif newStatus =='paused':
      # Just entered into 'paused' from some another status
      # Show message when status changed not from 'none' to 'paused'
      if lastStatus != 'none':
        sendmessage(_('Yandex.Disk'), _('Synchronization is paused'))
    elif newStatus == 'none':
      # Just entered into 'none' from some another status
      # Change menu sensitivity as daemon not started
      menu_YD_daemon_start.set_sensitive(True)
      menu_YD_daemon_stop.set_sensitive(False)
      sendmessage(_('Yandex.Disk'), _('Yandex.Disk daemon is stopped'))
    else:                          # newStatus = 'error'
      # Just entered into 'error' from some another status
      sendmessage(_('Yandex.Disk'), _('Synchronization ERROR'))
    if lastStatus == 'none':
      # When 'none' changed to something else (probably not to 'error' :) )
      # then it means that daemon has just started.
      # Change menu sensitivity
      menu_YD_daemon_start.set_sensitive(False)
      menu_YD_daemon_stop.set_sensitive(True)
      sendmessage(_('Yandex.Disk'), _('Yandex.Disk daemon is started'))
    lastStatus = newStatus         # remember last status
  
  # Handle timer time-outs 
  if triggeredByWatcher:           # True means that it is called by iNonify watcher
    # Recreate watcher timer.
    # It delays the timer triggering on 2 sec after last triggering from watcher.
    stopTimer(watchTimer)
    watchTimer = GLib.timeout_add_seconds(2, handleEvent, False)
    #debugPrint('Watch timer recreated, delay=2')
    timerTriggeredCount = 0        # reset counter as it was triggered by watcher
  else:
    if newStatus != 'busy':        # in 'busy' keep 2 sec update interval
      if timerTriggeredCount < 8:  # increase interval up to 10 sec (3+7)
        stopTimer(watchTimer)
        delaytime = 3 + timerTriggeredCount # increase delay for next time
        #debugPrint('Watch timer recreated, delay=%d' % delaytime)
        watchTimer = GLib.timeout_add_seconds(delaytime, handleEvent, False)
        timerTriggeredCount += 1
  return True                      # To continue activations by timer.

def updateIconTheme():
  global iconThemePath
  global ind
  global icon_busy
  global icon_idle
  global icon_pause
  global icon_error
  # Read defaults from dconf Settings
  try:
    if settings.get_boolean("theme"):
      themeStyle = 'light'
    else:
      themeStyle = 'dark'
  except:
      themeStyle = 'dark'
  # Set appropriate paths to icons
  iconThemePath = os.path.join(installDir, 'icons', themeStyle )
  icon_busy =  os.path.join(iconThemePath, 'yd-busy1.png')
  icon_idle =  os.path.join(iconThemePath, 'yd-ind-idle.png')
  icon_pause = os.path.join(iconThemePath, 'yd-ind-pause.png')
  icon_error = os.path.join(iconThemePath, 'yd-ind-error.png')

def updateIcon():
  # Change indicator icon according to the current status
  global newStatus
  global lastStatus
  global icon_busy
  global icon_idle
  global icon_pause
  global icon_error
  global iconAnimationTimer
  global seqNum
  #debugPrint('Icon updating:      Status: %s -> %s' % (lastStatus, newStatus))
  if newStatus == 'busy' and lastStatus != 'busy':
    # Just entered into 'busy' status
    ind.set_icon(icon_busy)
    seqNum = 1           # start animation from first icon
    # Create animation sequence timer
    iconAnimationTimer = GLib.timeout_add(777, iconAnimation, 'iconAnimation')
    return               # there is nothing to do any more here
  if newStatus != 'busy' and lastStatus == 'busy':
    # Just gone from 'busy' status
    # When 'busy' changed to something else then stop icon animation
    stopTimer(iconAnimationTimer)
  if newStatus == 'idle':
    ind.set_icon(icon_idle)
  elif newStatus == 'error':
    ind.set_icon(icon_error)
  else:                  # newStatus is 'none' or 'paused'
    ind.set_icon(icon_pause)

def iconAnimation(widget):
  # Triggered by iconAnimationTimer timer and changes busy icon by loop
  global seqNum
  global ind
  global iconThemePath
  seqFile = 'yd-busy'+str(seqNum)+'.png'
  ind.set_icon(os.path.join(iconThemePath, seqFile))
  if seqNum < 5:         # 5 icons in loop (1-2-3-4-5-1-2-3...)
    seqNum += 1
  else:
    seqNum = 1
  return True            # True required to continue triggering by timer

def activateActions(activate):
  # Actions for nautilus
  ret = subprocess.call(["dpkg -s nautilus>/dev/null 2>&1"], shell = True)
  debugPrint("Nautilus installed: %s"%str(ret == 0))
  if ret == 0:
    # Install/remove actions for Nautilus
    #debugPrint("Install/remove actions for Nautilus")
    ver = subprocess.check_output(["lsb_release -r | sed -n '1{s/[^0-9]//g;p;q}'"], shell = True)
    #debugPrint(int(ver))
    if int(ver) > 1210:
      nautilusPath = ".local/share/nautilus/scripts"
    else:
      nautilusPath = ".gnome2/nautilus-scripts/"
    debugPrint(nautilusPath)
    if activate:
      copyFile(os.path.join(installDir,"fm-actions/Nautilus_Nemo/publish"), os.path.join(userHome,nautilusPath,"Опубликовать через Yandex.Disk"))
      copyFile(os.path.join(installDir,"fm-actions/Nautilus_Nemo/unpublish"), os.path.join(userHome,nautilusPath,"Убрать из публикации через Yandex.disk"))
    else:
      deleteFile(os.path.join(userHome,nautilusPath,"Опубликовать через Yandex.Disk"))
      deleteFile(os.path.join(userHome,nautilusPath,"Убрать из публикации через Yandex.disk"))
  # Actions for Nemo
  ret = subprocess.call(["dpkg -s nemo>/dev/null 2>&1"], shell = True)
  debugPrint("Nemo installed: %s"%str(ret == 0))
  if ret == 0:
    # Install/remove actions for Nemo
    #debugPrint("Install/remove actions for Nemo")
    if activate:
      copyFile(os.path.join(installDir,"fm-actions/Nautilus_Nemo/publish"), os.path.join(userHome,".gnome2/nemo-scripts","Опубликовать через Yandex.Disk"))
      copyFile(os.path.join(installDir,"fm-actions/Nautilus_Nemo/unpublish"), os.path.join(userHome,".gnome2/nemo-scripts","Убрать из публикации через Yandex.disk"))
    else:
      deleteFile(os.path.join(userHome,".gnome2/nemo-scripts", "Опубликовать через Yandex.Disk"))
      deleteFile(os.path.join(userHome,".gnome2/nemo-scripts", "Убрать из публикации через Yandex.disk"))
  # Actions for Thunar
  ret = subprocess.call(["dpkg -s thunar>/dev/null 2>&1"], shell = True)
  debugPrint("Thunar installed: %s"%str(ret == 0))
  if ret == 0:
    # Install/remove actions for Thunar
    if activate:
      if subprocess.call(["grep 'yandex-disk publish' "+os.path.join(userHome,".config/Thunar/uca.xml")+">/dev/null 2>&1"], shell=True) != 0:
        subprocess.call(["sed", "-i", 's/<\/actions>/<action><icon>folder-publicshare<\/icon><name>Публикация через Yandex.Disk<\/name><command>yandex-disk publish %f | xclip -filter -selection clipboard; zenity --info --window-icon=\/opt\/yd-tools\/icons\/yd-128.png --title="Yandex.Disk" --ok-label="Закрыть" --text="Ссылка на файл: %f скопирована в буфер обмена."<\/command><description><\/description><patterns>*<\/patterns><directories\/><audio-files\/><image-files\/><other-files\/><text-files\/><video-files\/><\/action><\/actions>/g', os.path.join(userHome,".config/Thunar/uca.xml") ])
      if subprocess.call(["grep 'yandex-disk unpublish' "+os.path.join(userHome,".config/Thunar/uca.xml")+">/dev/null 2>&1"], shell=True) != 0:
        subprocess.call(["sed", "-i", 's/<\/actions>/<action><icon>folder<\/icon><name>Убрать из публикации через Yandex.disk<\/name><command>zenity --info --window-icon=\/opt\/yd-tools\/icons\/yd-128_g.png --ok-label="Закрыть" --title="Yandex.Disk" --text="Убрать из публикации через Yandex.disk: \`yandex-disk unpublish %f\`"<\/command><description><\/description><patterns>*<\/patterns><directories\/><audio-files\/><image-files\/><other-files\/><text-files\/><video-files\/><\/action><\/actions>/g', os.path.join(userHome,".config/Thunar/uca.xml")])
    else:
      subprocess.call(["sed", "-i", "s/<action><icon>.*<\/icon><name>Публикация через Yandex.Disk.*<\/action>//", os.path.join(userHome,".config/Thunar/uca.xml") ])
      subprocess.call(["sed", "-i", "s/<action><icon>.*<\/icon><name>Убрать из публикации через Yandex.disk.*<\/action>//", os.path.join(userHome,".config/Thunar/uca.xml") ])
  # Actions for Dolphin
  ret = subprocess.call(["dpkg -s dolphin>/dev/null 2>&1"], shell = True)
  debugPrint("Dolphin installed: %s"%str(ret == 0))
  if ret == 0:
    # Install/remove actions for Dolphin
    if activate:
      copyFile(os.path.join(installDir,"fm-actions/Dolphin/publish.desktop"), os.path.join(userHome,".kde/share/kde4/services/publish.desktop"))
      copyFile(os.path.join(installDir,"fm-actions/Dolphin/unpublish.desktop"), os.path.join(userHome,".kde/share/kde4/services/unpublish.desktop"))
    else:
      deleteFile(os.path.join(userHome,".kde/share/kde4/services/publish.desktop"))
      deleteFile(os.path.join(userHome,".kde/share/kde4/services/unpublish.desktop"))

def optionsCheck():
  global daemonConfig
  roVal = False
  ovVal = False
  try:
    cfgFile = open(daemonConfig, 'rU')
  except:
    debugPrint('CFG File open error')
  else:  
    fileContent = cfgFile.read()
    cfgFile.close()
    roVal = 'read-only="' in fileContent
    ovVal = 'overwrite="' in fileContent
    settings.set_boolean('optionreadonly', roVal)
    settings.set_boolean('optionoverwrite', ovVal)

def optionsSave():
  global daemonConfig
  roExist = False
  ovExist = False
  roVal = settings.get_boolean('optionreadonly')
  ovVal = settings.get_boolean('optionoverwrite')
  try:
    cfgFile = open(daemonConfig, 'r')
  except:
    debugPrint('CFG File open error')
  else:  
    newFile, newConfig = tempfile.mkstemp()
    for line in cfgFile:
      if 'read-only="' in line:
        roExist = True
        if not roVal:
          continue
      elif 'overwrite="' in line:
        ovExist = True
        if not (roVal and ovVal) :
          continue
      os.write(newFile, line.encode('utf-8')) # encode required for Python3 
    if roVal and not roExist:
      os.write(newFile, b'read-only=""\n')    # encode required for Python3
    if ovVal and roVal and not ovExist:
      os.write(newFile, b'overwrite=""\n')    # encode required for Python3
    cfgFile.close()
    os.close(newFile)
    deleteFile(daemonConfig)
    copyFile(newConfig, daemonConfig)

###################### MAIN LOOP #########################
if __name__ == '__main__':
  ### Application variables and settings ###
  appVersion = '1.1.6'
  appName = 'yandex-disk-indicator'
  installDir = os.path.dirname(os.path.realpath(__file__))
  userHome = os.getenv("HOME")
  # Define .desktop files locations for auto-start facility
  autoStartSource = os.path.join(installDir,'Yandex.Disk-indicator.desktop')
  autoStartDestination = os.path.join(userHome, '.config', 'autostart', 'Yandex.Disk-indicator.desktop')
  autoStartSource1 = os.path.join(installDir,'Yandex.Disk.desktop')
  autoStartDestination1 = os.path.join(userHome, '.config', 'autostart', 'Yandex.Disk.desktop')
  # Yandex.Disk configuration file path
  daemonConfig = os.path.join(userHome, '.config', 'yandex-disk', 'config.cfg')
  # Set LANG environment for daemon output (it must be 'en' for correct parsing)
  os.putenv('LANG','en_US.UTF-8')

  ### Configuration settings ###
  try:
    settings = Gio.Settings.new("apps."+appName)
  except:
    sys.exit('dconf settings are not found!')
  # Read some settings to variables
  try:
    verboseDebug = settings.get_boolean("debug")
  except:
    verboseDebug = False
  try:
    notificationSetting = settings.get_boolean("notifications")
  except:
    notificationSetting = True
  # Update auto-start settings according to actual existence of files in ~/.config/autostart
  try:
    settings.set_boolean('autostart',os.path.isfile(autoStartDestination))
    settings.set_boolean('autostartdaemon',os.path.isfile(autoStartDestination1))
  except:
    sys.exit('dconf settings are not found!')
  
  ### Localization ###
  debugPrint("Current Locale : %s" % str(locale.getlocale()))
  # Try to load translation
  try:
    appTranslate = gettext.translation(appName, '/usr/share/locale', fallback = True)
    _ = appTranslate.gettext
  except:
    _ = str

  ### Activate FM actions according to settings if lock file not exists (probably it is a first run)
  lockFileName = '/tmp/'+appName+'.lock'
  if not os.path.isfile(lockFileName):
    try:
      extOn = settings.get_boolean("fmextensions")
    except:
      extOn = False
    activateActions(extOn)

  ### Check for already running instance of the application ###
  try:
    lockFile = open(lockFileName,'w')
    # Try to acquire lock
    fcntl.flock(lockFile, fcntl.LOCK_EX|fcntl.LOCK_NB)
    # File hasn't been locked before
  except:
    # File is already locked
    sys.exit(_('Yandex.Disk Indicator instance already running\n(file /tmp/%s.lock is locked by another process)')%appName)
  lockFile.write('%d\n'%os.getpid())
  lockFile.flush()
  
  ### Yandex.Disk daemon ###
  # Check that Yandex.Disk daemon is running
  while not checkDaemon():    # no running daemon ?
    # Try to start it
    err = startYDdaemon() 
    if checkDaemon() and (err == ''):
      break
    # If it fails to start then show dialogue
    elif daemonErrorDialog(err) != 0:
      appExit()               # something wrong. It's better to exit right now
  # Get current status. Daemon must response correctly at this point.
  checkDaemonStatus()
  # Determine the Yandex.Disk folder.
  yandexDiskFolder = getYDFolder()
  # Update daemon start option settings according to the current configuration in ~/.config/yandex-disk/config.cfg
  optionsCheck()
  
  ### Application Indicator ###
  ## Icons ##
  yandexDiskIcon = os.path.join(installDir,'icons', 'yd-128.png')
  # Paths to rest icons are defined in updateIconTheme
  updateIconTheme()
  # Define the icon animation timer variable
  iconAnimationTimer = 0
  ## Indicator creation ##
  ind = appindicator.Indicator.new("yandex-disk", icon_pause, appindicator.IndicatorCategory.APPLICATION_STATUS)
  ind.set_status (appindicator.IndicatorStatus.ACTIVE)
  lastStatus = 'paused'       # according to initial indicator icon
  ## Menu ##
  renderMenu()                # Prepare and attach menu to indicator
  
  ### On-screen notifications ###
  Notify.init(appName)        # Initialize notification engine
  notifier = Notify.Notification()
  
  ### Status file update watcher ###
  watchManager = pyinotify.WatchManager()
  watchFile = os.path.join(yandexDiskFolder,'.sync/cli.log')   # Watched item
  debugPrint('Watched file is: "%s"'%watchFile)
  # Event handler class for iNotifier
  class EventHandler(pyinotify.ProcessEvent):
    def process_IN_MODIFY(self, event):
      #debugPrint( "Modified: %s" %  os.path.join(event.path, event.name))
      handleEvent(True)       # True means that handler called by iNotify watcher
  # Create iNotifier
  iNotifier = pyinotify.AsyncNotifier(watchManager, EventHandler(), read_freq = 1)
  # Add watch
  watchManager.add_watch(watchFile, pyinotify.IN_MODIFY, rec = False)
  # Create thread for asynchronous iNotifier
  loop_thread = threading.Thread(target=asyncore.loop, name = "Asyncore Loop" )
  loop_thread.start()
  asyncTimer = GLib.timeout_add_seconds(1, joinTreads)
  
  ### Timer triggered event staff ###
  # Set timer to update icon/menu by the eventHandler in 1 second
  watchTimer = GLib.timeout_add_seconds(1, handleEvent, False)
  timerTriggeredCount = 0
  Gtk.main()                  # Start GTK Main
